@@ -1,10 +1,9 @@
 package Net::DNS::Packet;
 
-#
-# $Id: Packet.pm 1523 2016-12-29 16:53:52Z willem $
-#
-use vars qw($VERSION);
-$VERSION = (qw$LastChangedRevision: 1523 $)[1];
+use strict;
+use warnings;
+
+our $VERSION = (qw$Id: Packet.pm 1959 2024-01-17 08:55:01Z willem $)[2];
 
 
 =head1 NAME
@@ -15,7 +14,7 @@ Net::DNS::Packet - DNS protocol packet
 
     use Net::DNS::Packet;
 
-    $query = new Net::DNS::Packet( 'example.com', 'MX', 'IN' );
+    $query = Net::DNS::Packet->new( 'example.com', 'MX', 'IN' );
 
     $reply = $resolver->send( $query );
 
@@ -27,10 +26,10 @@ A Net::DNS::Packet object represents a DNS protocol packet.
 =cut
 
 
-use strict;
 use integer;
 use Carp;
 
+use Net::DNS::Parameters qw(:dsotype);
 use constant UDPSZ => 512;
 
 BEGIN {
@@ -44,10 +43,10 @@ BEGIN {
 
 =head2 new
 
-    $packet = new Net::DNS::Packet( 'example.com' );
-    $packet = new Net::DNS::Packet( 'example.com', 'MX', 'IN' );
+    $packet = Net::DNS::Packet->new( 'example.com' );
+    $packet = Net::DNS::Packet->new( 'example.com', 'MX', 'IN' );
 
-    $packet = new Net::DNS::Packet();
+    $packet = Net::DNS::Packet->new();
 
 If passed a domain, type, and class, new() creates a Net::DNS::Packet
 object which is suitable for making a DNS query for the specified
@@ -59,8 +58,8 @@ If called with an empty argument list, new() creates an empty packet.
 =cut
 
 sub new {
-	return &decode if ref $_[1];
-	my $class = shift;
+	my ( $class, @arg ) = @_;
+	return &decode if ref $arg[0];
 
 	my $self = bless {
 		status	   => 0,
@@ -70,7 +69,7 @@ sub new {
 		additional => [],
 		}, $class;
 
-	$self->{question} = [Net::DNS::Question->new(@_)] if scalar @_;
+	$self->{question} = [Net::DNS::Question->new(@arg)] if scalar @arg;
 
 	return $self;
 }
@@ -80,8 +79,9 @@ sub new {
 
 =pod
 
-    $packet = new Net::DNS::Packet( \$data );
-    $packet = new Net::DNS::Packet( \$data, 1 );	# debug
+    $packet = Net::DNS::Packet->decode( \$data );
+    $packet = Net::DNS::Packet->decode( \$data, 1 );	# debug
+    $packet = Net::DNS::Packet->new( \$data ... );
 
 If passed a reference to a scalar containing DNS packet data, a new
 packet object is created by decoding the data.
@@ -93,7 +93,7 @@ Decoding errors, including data corruption and truncation, are
 collected in the $@ ($EVAL_ERROR) variable.
 
 
-    ( $packet, $length ) = new Net::DNS::Packet( \$data );
+    ( $packet, $length ) = Net::DNS::Packet->decode( \$data );
 
 If called in array context, returns a packet object and the number
 of octets successfully decoded.
@@ -115,12 +115,12 @@ sub decode {
 	my $self;
 	eval {
 		local $SIG{__DIE__};
-		die 'corrupt wire-format data' if length($$data) < HEADER_LENGTH;
+		my $length = length $$data;
+		die 'corrupt wire-format data' if $length < HEADER_LENGTH;
 
 		# header section
 		my ( $id, $status, @count ) = unpack 'n6', $$data;
 		my ( $qd, $an, $ns, $ar ) = @count;
-		$offset = HEADER_LENGTH;
 
 		$self = bless {
 			id	   => $id,
@@ -130,40 +130,50 @@ sub decode {
 			answer	   => [],
 			authority  => [],
 			additional => [],
-			answersize => length $$data
+			replysize  => $length
 			}, $class;
 
 		# question/zone section
 		my $hash = {};
 		my $record;
+		$offset = HEADER_LENGTH;
 		while ( $qd-- ) {
-			( $record, $offset ) = decode Net::DNS::Question( $data, $offset, $hash );
+			( $record, $offset ) = Net::DNS::Question->decode( $data, $offset, $hash );
 			CORE::push( @{$self->{question}}, $record );
 		}
 
 		# RR sections
 		while ( $an-- ) {
-			( $record, $offset ) = decode Net::DNS::RR( $data, $offset, $hash );
+			( $record, $offset ) = Net::DNS::RR->decode( $data, $offset, $hash );
 			CORE::push( @{$self->{answer}}, $record );
 		}
 
 		while ( $ns-- ) {
-			( $record, $offset ) = decode Net::DNS::RR( $data, $offset, $hash );
+			( $record, $offset ) = Net::DNS::RR->decode( $data, $offset, $hash );
 			CORE::push( @{$self->{authority}}, $record );
 		}
 
 		while ( $ar-- ) {
-			( $record, $offset ) = decode Net::DNS::RR( $data, $offset, $hash );
+			( $record, $offset ) = Net::DNS::RR->decode( $data, $offset, $hash );
 			CORE::push( @{$self->{additional}}, $record );
 		}
 
-		return $self;
+		return unless $offset == HEADER_LENGTH;
+		return unless $self->header->opcode eq 'DSO';
+
+		$self->{dso} = [];
+		my $limit = $length - 4;
+		while ( $offset < $limit ) {
+			my ( $t, $l, $v ) = unpack "\@$offset n2a*", $$data;
+			CORE::push( @{$self->{dso}}, [$t, substr( $v, 0, $l )] );
+			$offset += ( $l + 4 );
+		}
 	};
 
 	if ($debug) {
 		local $@ = $@;
 		print $@ if $@;
-		$self->print if $self;
+		eval { $self->print };
 	}
 
 	return wantarray ? ( $self, $offset ) : $self;
@@ -183,25 +193,25 @@ Truncation may be specified using a non-zero optional size argument.
 =cut
 
 sub data {
-	&encode;
+	return &encode;
 }
 
 sub encode {
 	my ( $self, $size ) = @_;				# uncoverable pod
 
 	my $edns = $self->edns;					# EDNS support
-	my @addl = grep !$_->isa('Net::DNS::RR::OPT'), @{$self->{additional}};
-	$self->{additional} = [@addl, $edns] if $edns->_specified;
+	my @addl = grep { !$_->isa('Net::DNS::RR::OPT') } @{$self->{additional}};
+	$self->{additional} = [$edns, @addl] if $edns->_specified;
 
 	return $self->truncate($size) if $size;
 
 	my @part = qw(question answer authority additional);
-	my @size = map scalar( @{$self->{$_}} ), @part;
+	my @size = map { scalar @{$self->{$_}} } @part;
 	my $data = pack 'n6', $self->header->id, $self->{status}, @size;
 	$self->{count} = [];
 
 	my $hash = {};						# packet body
-	foreach my $component ( map @{$self->{$_}}, @part ) {
+	foreach my $component ( map { @{$self->{$_}} } @part ) {
 		$data .= $component->encode( length $data, $hash, $self );
 	}
 
@@ -220,7 +230,7 @@ represents the header section of the packet.
 
 sub header {
 	my $self = shift;
-	bless \$self, q(Net::DNS::Header);
+	return bless \$self, q(Net::DNS::Header);
 }
 
 
@@ -238,8 +248,8 @@ extension OPT RR.
 sub edns {
 	my $self = shift;
 	my $link = \$self->{xedns};
-	($$link) = grep $_->isa(qw(Net::DNS::RR::OPT)), @{$self->{additional}} unless $$link;
-	$$link = new Net::DNS::RR( type => 'OPT' ) unless $$link;
+	($$link) = grep { $_->isa(qw(Net::DNS::RR::OPT)) } @{$self->{additional}} unless $$link;
+	$$link = Net::DNS::RR->new( type => 'OPT' ) unless $$link;
 	return $$link;
 }
 
@@ -257,12 +267,11 @@ response to an EDNS query.
 =cut
 
 sub reply {
-	my $query  = shift;
-	my $UDPmax = shift;
+	my ( $query, @UDPmax ) = @_;
 	my $qheadr = $query->header;
 	croak 'erroneous qr flag in query packet' if $qheadr->qr;
 
-	my $reply  = new Net::DNS::Packet();
+	my $reply  = Net::DNS::Packet->new();
 	my $header = $reply->header;
 	$header->qr(1);						# reply with same id, opcode and question
 	$header->id( $qheadr->id );
@@ -275,11 +284,11 @@ sub reply {
 	$header->rd( $qheadr->rd );				# copy these flags into reply
 	$header->cd( $qheadr->cd );
 
-	return $reply unless grep $_->isa('Net::DNS::RR::OPT'), @{$query->{additional}};
+	return $reply unless grep { $_->isa('Net::DNS::RR::OPT') } @{$query->{additional}};
 
 	my $edns = $reply->edns();
 	CORE::push( @{$reply->{additional}}, $edns );
-	$edns->size($UDPmax);
+	$edns->udpsize(@UDPmax);
 	return $reply;
 }
 
@@ -298,9 +307,10 @@ specifies the DNS zone to be updated.
 
 sub question {
 	my @qr = @{shift->{question}};
+	return @qr;
 }
 
-sub zone {&question}
+sub zone { return &question }
 
 
 =head2 answer, pre, prerequisite
@@ -318,10 +328,11 @@ not preexist.
 
 sub answer {
 	my @rr = @{shift->{answer}};
+	return @rr;
 }
 
-sub pre		 {&answer}
-sub prerequisite {&answer}
+sub pre		 { return &answer }
+sub prerequisite { return &answer }
 
 
 =head2 authority, update
@@ -338,9 +349,10 @@ specifies the RRs or RRsets to be added or deleted.
 
 sub authority {
 	my @rr = @{shift->{authority}};
+	return @rr;
 }
 
-sub update {&authority}
+sub update { return &authority }
 
 
 =head2 additional
@@ -354,6 +366,7 @@ section of the packet.
 
 sub additional {
 	my @rr = @{shift->{additional}};
+	return @rr;
 }
 
 
@@ -361,12 +374,15 @@ sub additional {
 
     $packet->print;
 
-Prints the packet data on the standard output in an ASCII format
-similar to that used in DNS zone files.
+Prints the entire packet to the currently selected output filehandle
+using the master file format mandated by RFC1035.
 
 =cut
 
-sub print { print &string; }
+sub print {
+	print &string;
+	return;
+}
 
 
 =head2 string
@@ -381,72 +397,86 @@ sub string {
 	my $self = shift;
 
 	my $header = $self->header;
-	my $update = $header->opcode eq 'UPDATE';
+	my $opcode = $header->opcode;
+	my $packet = $header->qr ? 'Response' : 'Query';
+	my $server = $self->{replyfrom};
+	my $length = $self->{replysize};
+	my $origin = $server ? ";; $packet received from [$server] $length octets\n" : "";
+	my @record = ( "$origin;; HEADER SECTION", $header->string );
 
-	my $server = $self->{answerfrom};
-	my $length = $self->{answersize};
-	my $string = $server ? ";; Answer received from $server ($length bytes)\n" : "";
+	if ( $opcode eq 'DSO' ) {
+		CORE::push( @record, ";; DSO SECTION" );
+		foreach ( @{$self->{dso}} ) {
+			my ( $t, $v ) = @$_;
+			CORE::push( @record, sprintf( ";;\t%s\t%s", dsotypebyval($t), unpack( 'H*', $v ) ) );
+		}
+		return join "\n", @record, "\n";
+	}
 
-	$string .= ";; HEADER SECTION\n" . $header->string;
+	my $edns = $self->edns;
+	CORE::push( @record, $edns->string ) if $edns->_specified;
 
-	my $question = $update ? 'ZONE' : 'QUESTION';
-	my @question = map $_->string, $self->question;
+	my @section  = $opcode eq 'UPDATE' ? qw(ZONE PREREQUISITE UPDATE) : qw(QUESTION ANSWER AUTHORITY);
+	my @question = $self->question;
 	my $qdcount  = scalar @question;
 	my $qds	     = $qdcount != 1 ? 's' : '';
-	$string .= join "\n;; ", "\n;; $question SECTION ($qdcount record$qds)", @question;
+	CORE::push( @record, ";; $section[0] SECTION ($qdcount record$qds)", map { ';; ' . $_->string } @question );
 
-	my $answer = $update ? 'PREREQUISITE' : 'ANSWER';
-	my @answer  = map $_->string, $self->answer;
+	my @answer  = $self->answer;
 	my $ancount = scalar @answer;
 	my $ans	    = $ancount != 1 ? 's' : '';
-	$string .= join "\n", "\n\n;; $answer SECTION ($ancount record$ans)", @answer;
+	CORE::push( @record, "\n;; $section[1] SECTION ($ancount record$ans)", map { $_->string } @answer );
 
-	my $authority = $update ? 'UPDATE' : 'AUTHORITY';
-	my @authority = map $_->string, $self->authority;
+	my @authority = $self->authority;
 	my $nscount   = scalar @authority;
 	my $nss	      = $nscount != 1 ? 's' : '';
-	$string .= join "\n", "\n\n;; $authority SECTION ($nscount record$nss)", @authority;
+	CORE::push( @record, "\n;; $section[2] SECTION ($nscount record$nss)", map { $_->string } @authority );
 
-	my @additional = map $_->string, $self->additional;
+	my @additional = $self->additional;
 	my $arcount    = scalar @additional;
 	my $ars	       = $arcount != 1 ? 's' : '';
-	$string .= join "\n", "\n\n;; ADDITIONAL SECTION ($arcount record$ars)", @additional;
+	my $EDNSmarker = join ' ', qq[;; {\t"EDNS-VERSION":], $edns->version, qq[}];
+	CORE::push( @record, "\n;; ADDITIONAL SECTION ($arcount record$ars)" );
+	CORE::push( @record, map { ( $_ eq $edns ) ? $EDNSmarker : $_->string } @additional );
 
-	return "$string\n\n";
+	return join "\n", @record, "\n";
 }
 
 
-=head2 answerfrom
+=head2 from
 
-    print "packet received from ", $packet->answerfrom, "\n";
+    print "packet received from ", $packet->from, "\n";
 
 Returns the IP address from which this packet was received.
-User-created packets will return undef for this method.
+This method will return undef for user-created packets.
 
 =cut
 
-sub answerfrom {
-	my $self = shift;
-
-	$self->{answerfrom} = shift if scalar @_;
-	$self->{answerfrom};
+sub from {
+	my ( $self, @argument ) = @_;
+	for (@argument) { $self->{replyfrom} = $_ }
+	return $self->{replyfrom};
 }
 
+sub answerfrom { return &from; }				# uncoverable pod
 
-=head2 answersize
 
-    print "packet size: ", $packet->answersize, " bytes\n";
+=head2 size
 
-Returns the size of the packet in bytes as it was received from a
-nameserver.  User-created packets will return undef for this method
+    print "packet size: ", $packet->size, " octets\n";
+
+Returns the size of the packet in octets as it was received from a
+nameserver.  This method will return undef for user-created packets
 (use length($packet->data) instead).
 
 =cut
 
-sub answersize {
-	shift->{answersize};
+sub size {
+	return shift->{replysize};
 }
 
+sub answersize { return &size; }				# uncoverable pod
+
 
 =head2 push
 
@@ -466,9 +496,9 @@ Section names may be abbreviated to the first three characters.
 =cut
 
 sub push {
-	my $self = shift;
-	my $list = $self->_section(shift);
-	CORE::push( @$list, grep ref($_), @_ );
+	my ( $self, $section, @rr ) = @_;
+	my $list = $self->_section($section);
+	return CORE::push( @$list, @rr );
 }
 
 
@@ -491,13 +521,11 @@ Section names may be abbreviated to the first three characters.
 =cut
 
 sub unique_push {
-	my $self = shift;
-	my $list = $self->_section(shift);
-	my @rr	 = grep ref($_), @_;
+	my ( $self, $section, @rr ) = @_;
+	my $list = $self->_section($section);
 
 	my %unique = map { ( bless( {%$_, ttl => 0}, ref $_ )->canonical => $_ ) } @rr, @$list;
-
-	scalar( @$list = values %unique );
+	return scalar( @$list = values %unique );
 }
 
 
@@ -514,7 +542,7 @@ Removes a single RR from the specified section of the packet.
 sub pop {
 	my $self = shift;
 	my $list = $self->_section(shift);
-	CORE::pop(@$list);
+	return CORE::pop(@$list);
 }
 
 
@@ -530,7 +558,7 @@ sub _section {				## returns array reference for section
 	my $self = shift;
 	my $name = shift;
 	my $list = $_section{unpack 'a3', $name} || $name;
-	$self->{$list} ||= [];
+	return $self->{$list} ||= [];
 }
 
 
@@ -539,7 +567,7 @@ sub _section {				## returns array reference for section
     $query = Net::DNS::Packet->new( 'www.example.com', 'A' );
 
     $query->sign_tsig(
-		'Khmac-sha512.example.+165+01018.private',
+		$keyfile,
 		fudge => 60
 		);
 
@@ -573,12 +601,6 @@ specified key.
     $query->sign_tsig( $tsig );
 
 
-The historical simplified syntax is still available, but additional
-options can not be specified.
-
-    $packet->sign_tsig( $key_name, $key );
-
-
 The response to an inbound request is signed by presenting the request
 in place of the key parameter.
 
@@ -603,17 +625,14 @@ does not support the suppressed signature scheme described in RFC2845.
 =cut
 
 sub sign_tsig {
-	my $self = shift;
-
-	eval {
+	my ( $self, @argument ) = @_;
+	return eval {
 		local $SIG{__DIE__};
 		require Net::DNS::RR::TSIG;
-		my $tsig = Net::DNS::RR::TSIG->create(@_);
+		my $tsig = Net::DNS::RR::TSIG->create(@argument);
 		$self->push( 'additional' => $tsig );
 		return $tsig;
-	} || do {
-		croak "$@\nTSIG: unable to sign packet";
-	};
+	} || return croak "$@\nTSIG: unable to sign packet";
 }
 
 
@@ -638,16 +657,13 @@ on the final packet in the absence of more specific information.
 =cut
 
 sub verify {
-	my $self = shift;
-
+	my ( $self, @argument ) = @_;
 	my $sig = $self->sigrr;
-	return $sig ? $sig->verify( $self, @_ ) : shift;
+	return $sig ? $sig->verify( $self, @argument ) : shift @argument;
 }
 
 sub verifyerr {
-	my $self = shift;
-
-	my $sig = $self->sigrr;
+	my $sig = shift->sigrr;
 	return $sig ? $sig->vrfyerrstr : 'not signed';
 }
 
@@ -659,7 +675,7 @@ The requisite cryptographic components are not integrated into
 Net::DNS but reside in the Net::DNS::SEC distribution available
 from CPAN.
 
-    $update = new Net::DNS::Update('example.com');
+    $update = Net::DNS::Update->new('example.com');
     $update->push( update => rr_add('foo.example.com A 10.1.2.3'));
     $update->sign_sig0('Kexample.com+003+25317.private');
 
@@ -679,23 +695,21 @@ sub sign_sig0 {
 	my $self = shift;
 	my $karg = shift;
 
-	eval {
+	return eval {
 		local $SIG{__DIE__};
-		require Net::DNS::RR::SIG;
 
 		my $sig0;
 		if ( ref($karg) eq 'Net::DNS::RR::SIG' ) {
 			$sig0 = $karg;
 
 		} else {
+			require Net::DNS::RR::SIG;
 			$sig0 = Net::DNS::RR::SIG->create( '', $karg );
 		}
 
 		$self->push( 'additional' => $sig0 );
 		return $sig0;
-	} || do {
-		croak "$@\nSIG0: unable to sign packet";
-	};
+	} || return croak "$@\nSIG0: unable to sign packet";
 }
 
 
@@ -712,10 +726,11 @@ sub sigrr {
 	my $self = shift;
 
 	my ($sig) = reverse $self->additional;
-	return undef unless $sig;
-	return $sig if $sig->type eq 'TSIG';
-	return $sig if $sig->type eq 'SIG';
-	return undef;
+	return unless $sig;
+	for ( $sig->type ) {
+		return $sig if /TSIG|SIG/;
+	}
+	return;
 }
 
 
@@ -727,7 +742,7 @@ The truncate method takes a maximum length as argument and then tries
 to truncate the packet and set the TC bit according to the rules of
 RFC2181 Section 9.
 
-The minimum maximum length that is honoured is 512 octets.
+The smallest length limit that is honoured is 512 octets.
 
 =cut
 
@@ -760,13 +775,12 @@ sub truncate {
 	$size = UDPSZ unless $size > UDPSZ;
 	$size -= $sigrr->_size if $sigrr;
 
-	my $data = pack 'x' x 12;				# header placeholder
-	my $hdsz = length $data;
+	my $data = pack 'x' x HEADER_LENGTH;			# header placeholder
 	$self->{count} = [];
 
 	my $tc;
 	my $hash = {};
-	foreach my $section ( map $self->{$_}, qw(question answer authority) ) {
+	foreach my $section ( map { $self->{$_} } qw(question answer authority) ) {
 		my @list;
 		foreach my $item (@$section) {
 			my $component = $item->encode( length $data, $hash );
@@ -782,7 +796,7 @@ sub truncate {
 
 	my %rrset;
 	my @order;
-	foreach my $item ( grep ref($_) ne ref($sigrr), $self->additional ) {
+	foreach my $item ( grep { ref($_) ne ref($sigrr) } $self->additional ) {
 		my $name  = $item->{owner}->canonical;
 		my $class = $item->{class} || 0;
 		my $key	  = pack 'nna*', $class, $item->{type}, $name;
@@ -809,18 +823,21 @@ sub truncate {
 	$self->{'additional'} = \@list;
 
 	my @part = qw(question answer authority additional);
-	my @size = map scalar( @{$self->{$_}} ), @part;
-	pack 'n6 a*', $self->header->id, $self->{status}, @size, substr( $data, $hdsz );
+	my @size = map { scalar @{$self->{$_}} } @part;
+	return pack 'n6 a*', $self->header->id, $self->{status}, @size, substr( $data, HEADER_LENGTH );
 }
 
 
 ########################################
 
 sub dump {				## print internal data structure
-	require Data::Dumper;					# uncoverable pod
+	my @data = @_;						# uncoverable pod
+	require Data::Dumper;
 	local $Data::Dumper::Maxdepth = $Data::Dumper::Maxdepth || 3;
 	local $Data::Dumper::Sortkeys = $Data::Dumper::Sortkeys || 1;
-	print Data::Dumper::Dumper(@_);
+	local $Data::Dumper::Useqq    = $Data::Dumper::Useqq	|| 1;
+	print Data::Dumper::Dumper(@data);
+	return;
 }
 
 
@@ -836,7 +853,7 @@ Portions Copyright (c)2002-2004 Chris Reinhardt.
 
 Portions Copyright (c)2002-2009 Olaf Kolkman
 
-Portions Copyright (c)2007-2015 Dick Franks
+Portions Copyright (c)2007-2019 Dick Franks
 
 All rights reserved.
 
@@ -845,7 +862,7 @@ All rights reserved.
 
 Permission to use, copy, modify, and distribute this software and its
 documentation for any purpose and without fee is hereby granted, provided
-that the above copyright notice appear in all copies and that both that
+that the original copyright notices appear in all copies and that both
 copyright notice and this permission notice appear in supporting
 documentation, and that the name of the author not be used in advertising
 or publicity pertaining to distribution of the software without specific
@@ -862,9 +879,11 @@ DEALINGS IN THE SOFTWARE.
 
 =head1 SEE ALSO
 
-L<perl>, L<Net::DNS>, L<Net::DNS::Update>, L<Net::DNS::Header>,
-L<Net::DNS::Question>, L<Net::DNS::RR>, L<Net::DNS::RR::TSIG>,
-RFC1035 Section 4.1, RFC2136 Section 2, RFC2845
+L<perl> L<Net::DNS> L<Net::DNS::Update> L<Net::DNS::Header>
+L<Net::DNS::Question> L<Net::DNS::RR> L<Net::DNS::RR::TSIG>
+L<RFC1035(4.1)|https://tools.ietf.org/html/rfc1035>
+L<RFC2136(2)|https://tools.ietf.org/html/rfc2136>
+L<RFC8945|https://tools.ietf.org/html/rfc8945>
 
 =cut
 
